function [I,alpha] = skm (X,k,sigma,varargin)
% Sparse Kernel Mean (SKM)
% See Efr\'en Cruz Cort\'es ans Clay Scott "Sparse Approximation of a
% Kernel Mean"
%
% When approximating a kernel mean of the form
%       (1/n)Sum_{all i} phi(.,xi),
% with a sparse version of the form
%       Sum_{i in I} alpha_i phi(.,xi),
% SKM finds the set I and coefficients alpha_i.
%
% This version only supports the Gaussian kernel, but can be altered to
% support other kernels.
%
% INPUT
% X: (nxd) data matrix
% k  = cardinality{I}: sparsity level
% sigma: kernel bandwidth
%
% OPTIONAL INPUT
% kernel: Only accepts 'gauss'. I have it to stress this fact. Modify
% according to your needs.
% H: Inner Product Space such that phi(.,x) in H. Supports L2 or phi's RKHS
% smplx: in {0,1}, indicates if alpha is to be projected into simplex
% amthd: "alpha method", if 0 it computes I with card{I}=k, and alpha with
% Matlab's pcg. If 1 it computes alpha with update version described in
% paper. If update method is used, the austoselection scheme can be used.
% auto: should be in {o,1}. Indicates if autoselection scheme is used. Only
% possible if amthd=1.
% epsilon: Stopping threshold for autoselection scheme.
%
% OUTPUT
% I: vector of indices of data points used for sparse approximation
% alpha: vector with nonzero coefficients in sparse approximation
%
% Efr\'en Cruz Cort\'es, 2015
%
n = size(X,1);

% - Default values of optional arguments -
kernel = 'gauss';
H = 'rkhs';
smplx = 0;
amthd = 1;
auto = 0;
epsilon = 10^-10;

% - Obtain optional arguments -
if nargin > 3
    for i=1:2:length(varargin)
        eval([varargin{i} '= varargin{i+1};']);
    end
end

% - H option for Gaussian kernel -
if strcmpi(kernel,'gauss')
    if strcmpi(H,'l2')
        sigma = sqrt(2)*sigma;
    end
else
    error('Modify code to support other kernels')
end

% - SKM -
first = randperm(n,1); % Pick first element of I at random
if amthd == 1
    [I,alpha] = autokca(X,k,first,sigma,epsilon,auto,smplx);
else
    if auto==1; error('For autoselection use alpha method 1'); end
    [I,YSd,SSd] = kca(X,k,first);
    alpha = getalpha(YSd,SSd,sigma,n,smplx);
end
%
end


function[I,YSdists,SSdists] = kca(X,k,first)
% k-center algorithm
% X is a matrix (nxd), interpret each row as d-dim point
% k is the model order
% vars:
% S will contain subset of X with nonzero coeffs
% Y = X\S
% returns I (index of S) and dist2(Y,S) and dist2(S,S), to be used later
n = size(X,1);
I = zeros(1,k); % indexes S
L = 1:n; % indexes Y
I(1) = first;
newx = X(first,:);
L(first) = [];

YSdists = zeros(n-1,k);
SSdists = zeros(k,k);
for i = 2:k
    YSdists(:,i-1) = dist2(X(L,:),newx);
    dys = min(YSdists(:,1:i-1),[],2);
    [~,wy] = max(dys,[],1);
    newx = X(L(wy),:);
    SSdists(i,1:i-1) = YSdists(wy,1:i-1);
    YSdists(wy,:)=[];
    I(i)=L(wy);
    L(wy)=[];
end
YSdists(:,k)=dist2(X(L,:),newx);
end


function [I,alpha] = autokca(X,k,first,sigma,epsilon,auto,smplx)
% k-center algorithm with autoselection scheme
% X is a matrix (nxd), interpret each row as d-dim point
% k is the model order
% vars:
% S will contain subset of X with nonzero coeffs
% Y = X\S
% returns I, index of S, and alpha coefficients
n = size(X,1);
I = zeros(1,k); % indexes S
L = 1:n; % indexes Y
I(1) = first;
newx = X(first,:);
L(first) = [];

YSdists = zeros(n-1,k);
SSdists = zeros(k,k);

gam = 1/(2*sigma^2); % used to evaluate kernel, specific to gaussian

K = zeros(k,k); % kernel matrix
H = zeros(k,k); % inverse of K
kappa = zeros(k,1);
alpha = cell(k,1);
z = zeros(k,1);
eh = zeros(k,1);
rh = zeros(k,1);

K(1,1) = exp(-gam * SSdists(1,1));
H(1,1) = 1/K(1,1);
YSdists(:,1) = dist2(X(L,:),newx);
kappa(1) = sum([exp(-gam*YSdists(:,1)); K(1,1)])/n;
alpha{1} = H(1,1)*kappa(1);
eh(1) = -alpha{1}*kappa(1);
for i = 2:k
    % --- I portion ---
    dys = min(YSdists(:,1:i-1),[],2);
    [~,wy] = max(dys,[],1);
    newx = X(L(wy),:);
    SSdists(i,1:i-1) = YSdists(wy,1:i-1);
    YSdists(wy,:)=[];
    I(i)=L(wy);
    L(wy)=[];
    
    YSdists(:,i) = dist2(X(L,:),newx);
    
    % --- alpha portion ---
    K(i,1:i) = exp(-gam * SSdists(i,1:i));
    K(1:i-1,i) = K(i,1:i-1)'; % b in the paper
    kappa(i) = sum([exp(-gam*YSdists(:,i)); K(1:i,i)])/n;
    q1 = H(1:i-1,1:i-1)*K(1:i-1,i);
    q0 = 1/(K(1,1) - K(i,1:i-1)*q1);
        % update: H(m+1)
        H(1:i,1:i) = [H(1:i-1,1:i-1) z(1:i-1); z(1:i-1)' 0] + ...
            q0 * [q1*q1' , -q1 ; -q1' , 1];
        % update: alpha(m+1)
        q2 = K(i,1:i-1)*alpha{i-1} - kappa(i);
        alpha{i} = [alpha{i-1} ; 0] + q0*[q2*q1 ; -q2];
    
    % --- error portion ---
    eh(i) = -alpha{i}'*kappa(1:i);
    rh(i) = abs((eh(i) - eh(i-1)) / (eh(i) - eh(1)));
    if auto == 1
        if rh(i)<epsilon; break; end % stop if error relatively small
    end
end
lastk = i;
I = I(1:lastk);
K = K(1:lastk,1:lastk);
kappa = kappa(1:lastk);
alpha = alpha{lastk};
if smplx == 1
    alpha = norm_alpha(alpha,K,kappa);
end
end


function [alpha] = getalpha(XCd,CCd,sigma,n,smplx)
% get K
CCd = CCd + triu(CCd.',1);
K = exp(-CCd / (2*sigma^2)); % specific for gaussian kernel
% get kappa
longK = exp(-XCd / (2*sigma^2)); % specific for gaussian kernel
kappa = sum([longK;K],1)'/n;
% compute alpha
alpha = pcg(K,kappa,10^-9,500); % I use pcg, but you could use something else
if smplx == 1
    alpha = norm_alpha(alpha,K,kappa);
end
end

function[alpha] = norm_alpha(alpha,K,kappa)
flag = true;
stepsize=.0001;
while flag
    alphap =alpha-2*(K*alpha-kappa)*stepsize; % Move in gradient direction
    alphap=simpproj(alphap); % Project to simplex
    if norm(alpha-alphap,1)<1e-4
        flag=false;
    end
    alpha=alphap;
end
end

function [ outvec] = simpproj( invec )
% SIMPPROJ input is a real vector, output is the vector projected onto the
% probabilistic simplex

% Based on "Efficient Projections onto the L1-Ball for Learning in High
% Dimensions", John Duchi, Shai Shalev-Shwartz, Yoram Singer, and Tushar Chandra

mu= sort(invec,'descend');
flag=true;
rho=1;
tally= mu(1);
while(true)
    if(rho==length(invec)) || (mu(rho+1)-(tally+mu(rho+1)-1)/(rho+1) <=0) ; break; end
    tally = tally+mu(rho+1);
    rho=rho+1;
end
theta =(tally-1)*rho^-1;
outvec= max(invec-theta,0);
end